"""
Representation of a Sonos Group.

Sonos players are always in groups, even if the group has only one player.
All players in a group play the same audio in synchrony.
Users can easily move players from one group to another without interrupting playback.
Transport controls, such as play, pause, skip to next track, and skip to previous track,
target groups rather than individual players.
Players must be part of the same household to be part of a group.

Reference: https://docs.sonos.com/docs/control
"""

from __future__ import annotations

import time
from contextlib import suppress
from typing import TYPE_CHECKING

from aiosonos.api.models import ContainerType, MetadataStatus, MusicService, PlayBackState
from aiosonos.api.models import GroupVolume as GroupVolumeData
from aiosonos.api.models import PlaybackStatus as PlaybackStatusData
from aiosonos.api.models import PlayModes as PlayModesData
from aiosonos.const import EventType, GroupEvent
from aiosonos.exceptions import FailedCommand

from .api.models import PlaybackActions as PlaybackActionsData

if TYPE_CHECKING:
    from aiosonos.api.models import Container, SessionStatus, Track

    from .api.models import Group as GroupData
    from .client import SonosLocalApiClient


class SonosGroup:
    """Representation of a Sonos Group."""

    def __init__(self, client: SonosLocalApiClient, data: GroupData) -> None:
        """Handle initialization."""
        self.client = client
        self.active_session_id: str | None = None
        self._data = data
        # to prevent race conditions at startup/init,
        # we set some default values for the status dicts here
        self._play_modes = PlayModes({})
        self._volume_data = GroupVolumeData(
            objectType="groupVolume", fixed=False, volume=0, mute=False
        )
        self._playback_actions = playback_actions = PlaybackActionsData(
            canCrossfade=False,
            canPause=False,
            canPlay=False,
            canSeek=False,
            canSkipBackward=False,
            canSkipForward=False,
            canStop=False,
        )
        self._playback_status_data: PlaybackStatusData = PlaybackStatusData(
            objectType="playbackStatus",
            availablePlaybackActions=playback_actions,
            isDucking=False,
            playbackState=PlayBackState.PLAYBACK_STATE_IDLE,
            playModes=PlayModesData(),
            positionMillis=0,
            previousPositionMillis=0,
        )
        self._playback_metadata_data: MetadataStatus = MetadataStatus(objectType="metadataStatus")
        self._playback_status_last_updated: float = 0.0
        self._unsubscribe_callbacks = []

    def cleanup(self) -> None:
        """Handle cleanup on deletion."""
        for unsubscribe_callback in self._unsubscribe_callbacks:
            unsubscribe_callback()
        self._unsubscribe_callbacks = []

    async def async_init(self) -> None:
        """Handle Async initialization."""
        # grab playback data and setup subscription
        try:
            self._volume_data = await self.client.api.group_volume.get_volume(self.id)
            self._playback_status_data = await self.client.api.playback.get_playback_status(self.id)
            self._playback_status_last_updated = time.time()
            self._playback_actions = PlaybackActions(
                self._playback_status_data["availablePlaybackActions"],
            )
            self._play_modes = PlayModes(self._playback_status_data["playModes"])
            self._playback_metadata_data = (
                await self.client.api.playback_metadata.get_metadata_status(
                    self.id,
                )
            )
            self._unsubscribe_callbacks = [
                await self.client.api.playback.subscribe(
                    self.id,
                    self._handle_playback_status_update,
                ),
                await self.client.api.group_volume.subscribe(
                    self.id,
                    self._handle_volume_update,
                ),
                await self.client.api.playback_metadata.subscribe(
                    self.id,
                    self._handle_metadata_status_update,
                ),
            ]
        except FailedCommand as err:
            if err.error_code == "groupCoordinatorChanged":
                # retrieving group details is not possible for remote groups when
                # connected to a player's local websocket.
                self._volume_data = {}
                self._playback_status_data = {}
                self._playback_actions = PlaybackActions({})
                self._play_modes = PlayModes({})
                self._playback_metadata_data = {}
                return
            raise
        else:
            self.client.signal_event(
                GroupEvent(
                    EventType.GROUP_ADDED,
                    self.id,
                    self,
                ),
            )

    @property
    def name(self) -> str:
        """Return the name of the group."""
        return self._data["name"]

    @property
    def id(self) -> str:
        """Return the group id."""
        return self._data["id"]

    @property
    def coordinator_id(self) -> str:
        """Return the coordinator's player Id (group leader)."""
        return self._data["coordinatorId"]

    @property
    def playback_state(self) -> PlayBackState:
        """Return the playback state of this group."""
        return (
            self._playback_status_data.get("playbackState")
            or self._data.get("playbackState")
            or PlayBackState.PLAYBACK_STATE_IDLE
        )

    @property
    def player_ids(self) -> list[str]:
        """Return the id's of this group's members."""
        return self._data["playerIds"]

    @property
    def area_ids(self) -> list[str]:
        """Return the area id's of this group (if any)."""
        return self._data.get("areaIds", [])

    @property
    def playback_actions(self) -> PlaybackActions:
        """Return the playback actions of this group."""
        return self._playback_actions

    @property
    def playback_metadata(self) -> MetadataStatus:
        """Return the MetadataStatus of this group."""
        return self._playback_metadata_data

    @property
    def position(self) -> float:
        """Return the (corrected) current position in (fractions of) seconds."""
        if self.playback_state == PlayBackState.PLAYBACK_STATE_PLAYING:
            return self._playback_status_data.get("positionMillis", 0) / 1000 + (
                time.time() - self._playback_status_last_updated
            )
        return self._playback_status_data.get("positionMillis", 0) / 1000

    @property
    def is_ducking(self) -> bool:
        """Return if the group's volume is currently ducked."""
        return self._playback_status_data.get("isDucking", False)

    @property
    def play_modes(self) -> PlayModes:
        """Return the play modes of this group."""
        return self._play_modes

    @property
    def container_type(self) -> ContainerType | None:
        """Return the container_type of the active source of this group (if any)."""
        if not (container := self._playback_metadata_data.get("container")):
            return None
        if container_type := container.get("type"):
            if container_type in ContainerType:
                return ContainerType(container_type)
            # return the raw string value if it's not a known container type
            return container_type
        return None

    @property
    def active_service(self) -> MusicService | None:
        """Return the active service of the active source of this group (if any)."""
        if not (container := self._playback_metadata_data.get("container")):
            return None
        if (container_id := container.get("id")) and (service_id := container_id.get("serviceId")):
            if service_id in MusicService:
                return MusicService(service_id)
            # return the raw string value if it's not a known container type
            return service_id
        return None

    async def play(self) -> None:
        """Send play command to group."""
        await self.client.api.playback.play(self.id)

    async def pause(self) -> None:
        """Send pause command to group."""
        await self.client.api.playback.pause(self.id)

    async def stop(self) -> None:
        """Send stop command to group."""
        try:
            # always first attempt to suspend the active session
            await self.client.api.playback_session.suspend(self.active_session_id)
        except FailedCommand:
            # this turns out the best way to clear the queue,
            # just send an invalid uri and it will clear the queue
            # and stop any playing content, including linein sources
            with suppress(FailedCommand):
                await self.play_stream_url("clear", None)
        finally:
            self.active_session_id = None

    async def toggle_play_pause(self) -> None:
        """Send play/pause command to group."""
        await self.client.api.playback.toggle_play_pause(self.id)

    async def skip_to_next_track(self) -> None:
        """Send skipToNextTrack command to group."""
        await self.client.api.playback.skip_to_next_track(self.id)

    async def skip_to_previous_track(self) -> None:
        """Send skipToPreviousTrack command to group."""
        await self.client.api.playback.skip_to_previous_track(self.id)

    async def set_play_modes(
        self,
        crossfade: bool | None = None,
        repeat: bool | None = None,
        repeat_one: bool | None = None,
        shuffle: bool | None = None,
    ) -> None:
        """Send setPlayModes command to group."""
        await self.client.api.playback.set_play_modes(
            self.id,
            crossfade=crossfade,
            repeat=repeat,
            repeat_one=repeat_one,
            shuffle=shuffle,
        )

    async def seek(self, position: int) -> None:
        """Send seek command to group."""
        await self.client.api.playback.seek(self.id, position)

    async def seek_relative(self, delta: int) -> None:
        """Send seekRelative command to group."""
        await self.client.api.playback.seek_relative(self.id, delta)

    async def load_line_in(
        self,
        device_id: str | None = None,
        play_on_completion: bool = False,  # noqa: FBT002
    ) -> None:
        """
        Send loadLineIn command to group.

        Parameters:
        - device_id (Optional): Represents the line-in source,
        any player in the household that supports line-in.
        The default value is the local deviceId.
        This is the same as the player ID returned in the player object.

        - play_on_completion (Optional): If true, start playback after
        loading the line-in source. If false, the player loads the cloud queue,
        but requires the play command to begin.
        If not provided, the default value is false.
        """
        await self.client.api.playback.load_line_in(
            self.id,
            device_id=device_id,
            play_on_completion=play_on_completion,
        )

    async def modify_group_members(
        self,
        player_ids_to_add: list[str],
        player_ids_to_remove: list[str],
    ) -> None:
        """Modify the group's members."""
        await self.client.api.groups.modify_group_members(
            self.id,
            player_ids_to_add,
            player_ids_to_remove,
        )

    async def set_group_members(
        self,
        player_ids: list[str],
        area_ids: list[str] | None = None,
    ) -> None:
        """Set/replace the group's members."""
        await self.client.api.groups.set_group_members(self.id, player_ids, area_ids)

    async def create_playback_session(
        self,
        app_id: str = "com.aiosonos.playback",
        app_context: str = "1",
        account_id: str | None = None,
        custom_data: dict | None = None,
    ) -> SessionStatus:
        """Create a new Playback Session."""
        session = await self.client.api.playback_session.create_session(
            self.id,
            app_id,
            app_context,
            account_id,
            custom_data,
        )
        self.active_session_id = session["sessionId"]
        return session

    async def play_stream_url(self, url: str, metadata: Container) -> None:
        """Create a new playback session and start playing a single (radio) stream URL."""
        await self.create_playback_session()
        await self.client.api.playback_session.load_stream_url(
            self.active_session_id,
            stream_url=url,
            play_on_completion=True,
            station_metadata=metadata,
        )

    async def play_cloud_queue(
        self,
        queue_base_url: str,
        http_authorization: str | None = None,
        use_http_authorization_for_media: bool | None = None,
        item_id: str | None = None,
        queue_version: str | None = None,
        position_millis: int | None = None,
        track_metadata: Track | None = None,
    ) -> None:
        """
        Create a new playback session and start playing a (cloud) queue.

        For all options, see:
        https://docs.sonos.com/reference/playbacksession-loadcloudqueue-sessionid
        """
        if not self.active_session_id:
            await self.create_playback_session()
        await self.client.api.playback_session.load_cloud_queue(
            self.active_session_id,
            queue_base_url=queue_base_url,
            http_authorization=http_authorization,
            use_http_authorization_for_media=use_http_authorization_for_media,
            item_id=item_id,
            queue_version=queue_version,
            position_millis=position_millis,
            play_on_completion=True,
            track_metadata=track_metadata,
        )

    def update_data(self, data: GroupData) -> None:
        """Update the player data."""
        if data == self._data:
            return
        for key, value in data.items():
            self._data[key] = value
        self.client.signal_event(
            GroupEvent(
                EventType.GROUP_UPDATED,
                data["id"],
                self,
            ),
        )

    def _handle_playback_status_update(self, data: PlaybackStatusData) -> None:
        """Handle playbackStatus update."""
        if data == self._playback_status_data:
            return
        self._playback_status_data = data
        self._playback_status_last_updated = time.time()
        self._playback_actions.raw_data.update(data["availablePlaybackActions"])
        self._play_modes.raw_data.update(data["playModes"])
        self.client.signal_event(
            GroupEvent(
                EventType.GROUP_UPDATED,
                self.id,
                self,
            ),
        )

    def _handle_metadata_status_update(self, data: MetadataStatus) -> None:
        """Handle MetadataStatus update."""
        if data == self._playback_metadata_data:
            return
        self._playback_metadata_data = data
        self.client.signal_event(
            GroupEvent(
                EventType.GROUP_UPDATED,
                self.id,
                self,
            ),
        )

    def _handle_volume_update(self, data: GroupVolumeData) -> None:
        """Handle volume update."""
        if data == self._volume_data:
            return
        self._volume_data = data
        self.client.signal_event(
            GroupEvent(
                EventType.GROUP_UPDATED,
                self.id,
                self,
            ),
        )


class PlaybackActions:
    """Representation of the PlaybackActions on a Sonos Group."""

    def __init__(self, raw_data: PlaybackActionsData) -> None:
        """Handle initialization."""
        self.raw_data = raw_data

    @property
    def can_skip_forward(self) -> bool:
        """Return if the group can skip forward."""
        return self.raw_data.get("canSkipForward", False)

    @property
    def can_skip_backward(self) -> bool:
        """Return if the group can skip backward."""
        return self.raw_data.get("canSkipBackward", False)

    @property
    def can_play(self) -> bool:
        """Return if the group can play."""
        return self.raw_data.get("canPlay", False)

    @property
    def can_pause(self) -> bool:
        """Return if the group can pause."""
        return self.raw_data.get("canPause", False)

    @property
    def can_stop(self) -> bool:
        """Return if the group can stop."""
        return self.raw_data.get("canStop", False)


class PlayModes:
    """Representation of the PlayModes on a Sonos Group."""

    def __init__(self, raw_data: PlayModesData) -> None:
        """Handle initialization."""
        self.raw_data = raw_data

    @property
    def crossfade(self) -> bool | None:
        """Return if crossfade is enabled."""
        return self.raw_data.get("crossfade")

    @property
    def repeat(self) -> bool | None:
        """Return if repeat is enabled."""
        return self.raw_data.get("repeat")

    @property
    def repeat_one(self) -> bool | None:
        """Return if repeat one is enabled."""
        return self.raw_data.get("repeatOne")

    @property
    def shuffle(self) -> bool | None:
        """Return if shuffle is enabled."""
        return self.raw_data.get("shuffle")



"""Utils/helpers for aiosonos."""

from __future__ import annotations

import ssl
from typing import TYPE_CHECKING

from aiosonos.const import LOCAL_API_TOKEN

if TYPE_CHECKING:
    from aiohttp import ClientSession

    from aiosonos.api.models import DiscoveryInfo


def create_ssl_context() -> ssl.SSLContext:
    """Create SSL context for Sonos local API connections."""
    ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    ctx.minimum_version = ssl.TLSVersion.TLSv1_2
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    return ctx


async def get_discovery_info(
    aiohttp_session: ClientSession,
    player_ip: str,
) -> DiscoveryInfo:
    """Get the discovery info for a player."""
    async with aiohttp_session.get(
        f"https://{player_ip}:1443/api/v1/players/local/info",
        headers={"X-Sonos-Api-Key": LOCAL_API_TOKEN},
        ssl=create_ssl_context(),
    ) as resp:
        resp.raise_for_status()
        discovery_info: DiscoveryInfo = await resp.json()
    return discovery_info



"""Constants/Models for the SonosApiClient."""

from __future__ import annotations

from dataclasses import dataclass
from enum import StrEnum
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .group import SonosGroup
    from .player import SonosPlayer

CLOUD_BASE_URL = "api.ws.sonos.com/control/api/v1"
LOCAL_API_TOKEN = "123e4567-e89b-12d3-a456-426655440000"  # noqa: S105
LOG_LEVEL_VERBOSE = 5


class EventType(StrEnum):
    """Event types for the Sonos Client."""

    GROUP_ADDED = "group_added"
    GROUP_UPDATED = "group_updated"
    GROUP_REMOVED = "group_removed"
    PLAYER_ADDED = "player_added"
    PLAYER_UPDATED = "player_updated"
    PLAYER_REMOVED = "player_removed"
    CONNECTED = "connected"
    DISCONNECTED = "disconnected"

    # special event type to match all events
    MATCH_ALL = "match_all"


@dataclass
class SonosEventBase:
    """Representation of an Event emitted in/by Sonos."""

    event_type: EventType
    object_id: str | None = None  # optional identifier such as player id or group id
    data: Any | None = None  # optional data


@dataclass
class PlayerEvent(SonosEventBase):
    """Representation of an Event emitted when a player was added, updated or removed."""

    data: SonosPlayer


@dataclass
class GroupEvent(SonosEventBase):
    """Representation of an Event emitted when a group was added, updated or removed."""

    data: SonosGroup


SonosEvent = PlayerEvent | GroupEvent



"""Handle AudioClip related endpoints for Sonos."""

from __future__ import annotations

from aiosonos.api.models import (
    AudioClip,
    AudioClipLEDBehavior,
    AudioClipPriority,
    AudioClipStatusEvent,
    AudioClipType,
)

from ._base import SonosNameSpace, SubscribeCallbackType, UnsubscribeCallbackType


class AudioClipNameSpace(SonosNameSpace):
    """AudioClip Namespace handlers."""

    namespace = "audioClip"
    event_type = "audioClip"
    _event_model = AudioClipStatusEvent
    _event_key = "playerId"

    async def load_audio_clip(
        self,
        player_id: str,
        name: str,
        app_id: str,
        stream_url: str | None = None,
        volume: int | None = None,
        priority: AudioClipPriority = AudioClipPriority.LOW,
        clip_type: AudioClipType = AudioClipType.CHIME,
        http_authorization: str | None = None,
        clip_led_behavior: AudioClipLEDBehavior = AudioClipLEDBehavior.NONE,
    ) -> AudioClip:
        """
        Send loadAudioClip command to player.

        Note that when connected to a local speaker's websocket,
        the player id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/audioclip-loadaudioclip-playerid
        """
        return await self.api.send_command(
            namespace=self.namespace,
            command="loadAudioClip",
            playerId=player_id,
            options={
                "name": name,
                "appId": app_id,
                "priority": priority,
                "clipType": clip_type,
                "streamUrl": stream_url,
                "httpAuthorization": http_authorization,
                "volume": volume,
                "clipLEDBehavior": clip_led_behavior,
            },
        )

    async def subscribe(
        self,
        player_id: str,
        callback: SubscribeCallbackType,
    ) -> UnsubscribeCallbackType:
        """
        Subscribe to events in the AudioClip namespace for given player.

        Returns handle to unsubscribe.

        Reference:
        https://docs.sonos.com/reference/audioclip-subscribe-playerid
        """
        return await self._handle_subscribe(player_id, callback)



"""Handle HomeTheater related endpoints for Sonos."""

from __future__ import annotations

from ._base import SonosNameSpace


class HomeTheaterNameSpace(SonosNameSpace):
    """HomeTheater Namespace handlers."""

    namespace = "homeTheater"
    event_type = "homeTheaterStatus"
    _event_model = None  # TODO: Add model
    _event_key = "playerId"

    async def load_home_theater_playback(
        self,
        player_id: str,
    ) -> None:
        """
        Send loadHomeTheaterPlayback command to player.

        Note that this command is not yet documented in the api docs and
        has been reverse engineered from the Sonos Web Controller.
        """
        return await self.api.send_command(
            namespace=self.namespace,
            command="loadHomeTheaterPlayback",
            playerId=player_id,
        )



"""Handle Playback related endpoints for Sonos."""

from __future__ import annotations

from aiosonos.api.models import LoadContentRequest, PlaybackStatus, PlayModes

from ._base import SonosNameSpace, SubscribeCallbackType, UnsubscribeCallbackType


class PlaybackNameSpace(SonosNameSpace):
    """PlaybackNameSpace Namespace handlers."""

    namespace = "playback"
    event_type = "playbackStatus"
    _event_model = PlaybackStatus
    _event_key = "groupId"

    async def get_playback_status(
        self,
        group_id: str,
    ) -> PlaybackStatus:
        """
        Send getPlaybackStatus command to group.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playback-getplaybackstatus-groupid
        """
        return await self.api.send_command(
            namespace=self.namespace,
            command="getPlaybackStatus",
            groupId=group_id,
        )

    async def load_line_in(
        self,
        group_id: str,
        device_id: str | None = None,
        play_on_completion: bool = False,  # noqa: FBT002
    ) -> None:
        """
        Send getPlaybackStatus command to group.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playback-loadlinein-groupid
        """
        options = {}
        if device_id is not None:
            options["deviceId"] = device_id
        if play_on_completion is not None:
            options["playOnCompletion"] = play_on_completion
        await self.api.send_command(
            namespace=self.namespace,
            command="loadLineIn",
            groupId=group_id,
            options=options,
        )

    async def load_content(
        self,
        group_id: str,
        content: LoadContentRequest,
    ) -> None:
        """Load content on the existing group's queue."""
        await self.api.send_command(
            namespace=self.namespace, command="loadContent", groupId=group_id, options=content
        )

    async def pause(
        self,
        group_id: str,
    ) -> None:
        """
        Send pause command to group.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playback-pause-groupid
        """
        return await self.api.send_command(
            namespace=self.namespace,
            command="pause",
            groupId=group_id,
        )

    async def play(
        self,
        group_id: str,
    ) -> None:
        """
        Send play command to group.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playback-play-groupid
        """
        return await self.api.send_command(
            namespace=self.namespace,
            command="play",
            groupId=group_id,
        )

    async def set_play_modes(
        self,
        group_id: str,
        repeat: bool | None = None,
        repeat_one: bool | None = None,
        shuffle: bool | None = None,
        crossfade: bool | None = None,
    ) -> None:
        """
        Send getPlaybackStatus command to group.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playback-setplaymodes-groupid
        """
        play_modes = PlayModes()
        if repeat is not None:
            play_modes["repeat"] = repeat
        if repeat_one is not None:
            play_modes["repeatOne"] = repeat_one
        if shuffle is not None:
            play_modes["shuffle"] = shuffle
        if crossfade is not None:
            play_modes["crossfade"] = crossfade
        await self.api.send_command(
            namespace=self.namespace,
            command="setPlayModes",
            groupId=group_id,
            options={
                "playModes": play_modes,
            },
        )

    async def seek(
        self,
        group_id: str,
        position_millis: int,
        item_id: str | None = None,
    ) -> None:
        """
        Send seek command to group.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playback-seek-groupid
        """
        options = {
            "positionMillis": position_millis,
        }
        if item_id is not None:
            options["itemId"] = item_id
        await self.api.send_command(
            namespace=self.namespace,
            command="seek",
            groupId=group_id,
            options=options,
        )

    async def seek_relative(
        self,
        group_id: str,
        delta_millis: int,
        item_id: str | None = None,
    ) -> None:
        """
        Send seekRelative command to group.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playback-seekrelative-groupid
        """
        options = {
            "deltaMillis": delta_millis,
        }
        if item_id is not None:
            options["itemId"] = item_id
        await self.api.send_command(
            namespace=self.namespace,
            command="seekRelative",
            groupId=group_id,
            options=options,
        )

    async def skip_to_next_track(
        self,
        group_id: str,
    ) -> None:
        """
        Send skipToNextTrack command to group.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playback-skiptonexttrack-groupid
        """
        return await self.api.send_command(
            namespace=self.namespace,
            command="skipToNextTrack",
            groupId=group_id,
        )

    async def skip_to_previous_track(
        self,
        group_id: str,
    ) -> None:
        """
        Send skipToPreviousTrack command to group.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playback-skiptoprevioustrack-groupid
        """
        return await self.api.send_command(
            namespace=self.namespace,
            command="skipToPreviousTrack",
            groupId=group_id,
        )

    async def toggle_play_pause(
        self,
        group_id: str,
    ) -> None:
        """
        Send togglePlayPause command to group.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playback-toggleplaypause-groupid
        """
        return await self.api.send_command(
            namespace=self.namespace,
            command="togglePlayPause",
            groupId=group_id,
        )

    async def subscribe(
        self,
        group_id: str,
        callback: SubscribeCallbackType,
    ) -> UnsubscribeCallbackType:
        """
        Subscribe to events in the PlaybackStatus namespace for given player.

        Returns handle to unsubscribe.

        Reference:
        https://docs.sonos.com/reference/playback-subscribe-groupid
        """
        return await self._handle_subscribe(group_id, callback)



"""Representation of a Sonos Player."""

from __future__ import annotations

from typing import TYPE_CHECKING

from aiosonos.api.models import AudioClipLEDBehavior, AudioClipType
from aiosonos.const import EventType, PlayerEvent

if TYPE_CHECKING:
    from aiosonos.group import SonosGroup

    from .api.models import Player as PlayerData
    from .api.models import PlayerVolume as PlayerVolumeData
    from .client import SonosLocalApiClient


class SonosPlayer:
    """Representation of a Sonos Player."""

    _active_group: SonosGroup | None

    def __init__(self, client: SonosLocalApiClient, data: PlayerData) -> None:
        """Handle initialization."""
        self.client = client
        self._data = data
        self._volume_data: PlayerVolumeData | None = None
        for group in client.groups:
            if group.coordinator_id == self.id or self.id in group.player_ids:
                self._active_group = group
                break
        else:
            self._active_group = None

    async def async_init(self) -> None:
        """Handle Async initialization."""
        # grab volume data and setup subscription
        self._volume_data = await self.client.api.player_volume.get_volume(self.id)
        await self.client.api.player_volume.subscribe(
            self.id,
            self._handle_volume_update,
        )

    @property
    def name(self) -> str:
        """Return the name of the player."""
        return self._data["name"]

    @property
    def id(self) -> str:
        """Return the player id."""
        return self._data["id"]

    @property
    def icon(self) -> str:
        """Return the icon."""
        return self._data.get("icon", "")

    @property
    def volume_level(self) -> int | None:
        """Return the current volume level of the player."""
        return self._volume_data.get("volume")

    @property
    def volume_muted(self) -> bool | None:
        """Return the current mute state of the player."""
        return self._volume_data.get("muted")

    @property
    def has_fixed_volume(self) -> bool | None:
        """Return if this player has a fixed volume level."""
        return self._volume_data.get("fixed")

    @property
    def group(self) -> SonosGroup | None:
        """Return the active group."""
        return self._active_group

    @property
    def is_coordinator(self) -> bool:
        """Return if this player is the coordinator of the active group it belongs to."""
        if not self.group:
            return False
        return self.group.coordinator_id == self.id

    @property
    def is_passive(self) -> bool:
        """Return if this player is the NOT a coordinator but a passive member of a group."""
        if not self.group:
            return False
        return self.group.coordinator_id != self.id

    @property
    def group_members(self) -> list[str]:
        """Return the player ids of the group members."""
        if not self.group:
            return []
        return self.group.player_ids

    async def set_volume(
        self,
        volume: int | None = None,
        muted: bool | None = None,
    ) -> None:
        """Set the volume of the player."""
        await self.client.api.player_volume.set_volume(self.id, volume, muted)

    async def duck(self, duration_millis: int | None = None) -> None:
        """Duck the volume of the player."""
        await self.client.api.player_volume.duck(self.id, duration_millis)

    async def leave_group(self) -> None:
        """Leave the active group this player is joined to (if any)."""
        if not self.group:
            return
        await self.client.api.groups.modify_group_members(
            self.group.id,
            player_ids_to_add=[],
            player_ids_to_remove=[self.id],
        )

    async def join_group(self, group_id: str) -> None:
        """Join a group."""
        await self.client.api.groups.modify_group_members(
            group_id,
            player_ids_to_add=[self.id],
            player_ids_to_remove=[],
        )

    async def play_audio_clip(
        self,
        url: str,
        volume: int | None = None,
        name: str | None = None,
    ) -> None:
        """Play an audio clip (announcement) on the player."""
        await self.client.api.audio_clip.load_audio_clip(
            self.id,
            name=name or "aiosonos",
            app_id="aiosonos",
            stream_url=url,
            volume=volume,
            clip_type=AudioClipType.CUSTOM,
            clip_led_behavior=AudioClipLEDBehavior.WHITE_LED_QUICK_BREATHING,
        )

    async def load_home_theater_playback(self) -> None:
        """Load Home Theater playback."""
        await self.client.api.home_theater.load_home_theater_playback(self.id)

    def update_data(self, data: PlayerData) -> None:
        """Update the player data."""
        self.check_active_group()
        if data == self._data:
            return
        for key, value in data.items():
            self._data[key] = value
        self.client.signal_event(
            PlayerEvent(
                EventType.PLAYER_UPDATED,
                data.get("id", self.id),
                self,
            ),
        )

    def check_active_group(self) -> None:
        """Check/set the active group of this player."""
        prev_group_id = self._active_group.id if self._active_group else None
        for group in self.client.groups:
            if group.coordinator_id == self.id or self.id in group.player_ids:
                self._active_group = group
                break
        if prev_group_id == self.group.id:
            return
        self.client.signal_event(
            PlayerEvent(
                EventType.PLAYER_UPDATED,
                self.id,
                self,
            ),
        )

    def _handle_volume_update(self, data: PlayerVolumeData) -> None:
        """Handle volume update."""
        if data == self._volume_data:
            return
        self._volume_data = data
        self.client.signal_event(
            PlayerEvent(
                EventType.PLAYER_UPDATED,
                self.id,
                self,
            ),
        )



"""Handle GroupVolume related endpoints for Sonos."""

from __future__ import annotations

from aiosonos.api.models import GroupVolume

from ._base import SonosNameSpace, SubscribeCallbackType, UnsubscribeCallbackType


class GroupVolumeNameSpace(SonosNameSpace):
    """GroupVolume Namespace handlers."""

    namespace = "groupVolume"
    event_type = "groupVolume"
    _event_model = GroupVolume
    _event_key = "groupId"

    async def set_volume(
        self,
        group_id: str,
        volume: int,
    ) -> None:
        """
        Send setVolume command to set the group's volume.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/groupvolume-setvolume-groupid
        """
        await self.api.send_command(
            namespace=self.namespace,
            command="setVolume",
            groupId=group_id,
            options={"volume": volume},
        )

    async def get_volume(
        self,
        group_id: str,
    ) -> GroupVolume:
        """
        Get the group's volume and mute state.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/groupvolume-getvolume-groupid
        """
        return await self.api.send_command(
            namespace=self.namespace,
            command="getVolume",
            groupId=group_id,
        )

    async def set_mute(
        self,
        group_id: str,
        muted: bool,
    ) -> None:
        """
        Send the setMute command to set the group's mute state.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/groupvolume-setmute-groupid
        """
        await self.api.send_command(
            namespace=self.namespace,
            command="setMute",
            groupId=group_id,
            options={"muted": muted},
        )

    async def set_relative_volume(
        self,
        group_id: str,
        volume_delta: int | None = None,
    ) -> None:
        """
        Send setRelativeVolume command to relatively change the group volume.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/groupvolume-setrelativevolume-groupid
        """
        await self.api.send_command(
            namespace=self.namespace,
            command="setRelativeVolume",
            groupId=group_id,
            options={"volumeDelta": volume_delta},
        )

    async def subscribe(
        self,
        group_id: str,
        callback: SubscribeCallbackType,
    ) -> UnsubscribeCallbackType:
        """
        Subscribe to events in the GroupVolume namespace for given group.

        Returns handle to unsubscribe.

        Reference:
        https://docs.sonos.com/reference/groupvolume-subscribe-groupid
        """
        return await self._handle_subscribe(group_id, callback)



"""Handle Groups related endpoints for Sonos."""

from __future__ import annotations

from aiosonos.api.models import GroupInfo, Groups

from ._base import SonosNameSpace, SubscribeCallbackType, UnsubscribeCallbackType


class GroupsNameSpace(SonosNameSpace):
    """Groups Namespace handlers."""

    namespace = "groups"
    event_type = "groups"
    _event_model = Groups
    _event_key = "householdId"

    async def modify_group_members(
        self,
        group_id: str,
        player_ids_to_add: list[str],
        player_ids_to_remove: list[str],
    ) -> GroupInfo:
        """
        Send modifyGroupMembers command to modify the group's members.

        Reference:
        https://docs.sonos.com/reference/groups-modifygroupmembers-groupid
        """
        return await self.api.send_command(
            namespace=self.namespace,
            command="modifyGroupMembers",
            groupId=group_id,
            options={
                "playerIdsToAdd": player_ids_to_add,
                "playerIdsToRemove": player_ids_to_remove,
            },
        )

    async def set_group_members(
        self,
        group_id: str,
        player_ids: list[str],
        area_ids: list[str] | None = None,
    ) -> GroupInfo:
        """
        Send setGroupMembers command to set/replace the group's members.

        Reference:
        https://docs.sonos.com/reference/groups-setgroupmembers-groupid
        """
        options = {
            "playerIds": player_ids,
        }
        if area_ids is not None:
            options["areaIds"] = area_ids
        return await self.api.send_command(
            namespace=self.namespace,
            command="setGroupMembers",
            groupId=group_id,
            options=options,
        )

    async def get_groups(
        self,
        household_id: str,
        include_device_info: bool = False,  # noqa: FBT002
    ) -> Groups:
        """
        Get all groups (and players) in a household.

        Reference:
        https://docs.sonos.com/reference/groups-getgroups-householdid
        """
        return await self.api.send_command(
            namespace=self.namespace,
            command="getGroups",
            householdId=household_id,
            options={
                "includeDeviceInfo": include_device_info,
            },
        )

    async def create_group(
        self,
        household_id: str,
        player_ids: list[str],
        music_context_group_id: str | None = None,
    ) -> GroupInfo:
        """
        Send createGroup command to create a new group.

        Reference:
        https://docs.sonos.com/reference/groups-creategroup-householdid
        """
        options = {
            "playerIds": player_ids,
        }
        if music_context_group_id is not None:
            options["musicContextGroupId"] = music_context_group_id
        return await self.api.send_command(
            namespace=self.namespace,
            command="createGroup",
            householdId=household_id,
            options=options,
        )

    async def subscribe(
        self,
        household_id: str,
        callback: SubscribeCallbackType,
    ) -> UnsubscribeCallbackType:
        """
        Subscribe to events in the Groups namespace for given player.

        Returns handle to unsubscribe.

        Reference:
        https://docs.sonos.com/reference/groups-subscribe-householdid
        """
        return await self._handle_subscribe(household_id, callback)



"""Handle PlaybackSession related endpoints for Sonos."""

from __future__ import annotations

from typing import TYPE_CHECKING

from aiosonos.api.models import SessionStatus, Track

from ._base import SonosNameSpace, SubscribeCallbackType, UnsubscribeCallbackType

if TYPE_CHECKING:
    from collections.abc import Container


class PlaybackSessionNameSpace(SonosNameSpace):
    """PlaybackSession Namespace handlers."""

    namespace = "playbackSession"
    event_type = "playbackSession"
    _event_model = SessionStatus
    _event_key = "sessionId"

    async def create_session(
        self,
        group_id: str,
        app_id: str,
        app_context: str,
        account_id: str | None = None,
        custom_data: dict | None = None,
    ) -> SessionStatus:
        """
        Send createSession command to group.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playbacksession-createsession-groupid
        """
        options = {
            "appId": app_id,
            "appContext": app_context,
        }
        if account_id is not None:
            options["accountId"] = account_id
        if custom_data is not None:
            options["customData"] = custom_data
        return await self.api.send_command(
            namespace=self.namespace,
            command="createSession",
            groupId=group_id,
            options=options,
        )

    async def load_cloud_queue(
        self,
        session_id: str,
        queue_base_url: str,
        http_authorization: str | None = None,
        use_http_authorization_for_media: bool | None = None,
        item_id: str | None = None,
        queue_version: str | None = None,
        position_millis: int | None = None,
        play_on_completion: bool | None = None,
        track_metadata: Track | None = None,
    ) -> None:
        """
        Send loadCloudQueue command to an (active) playback session).

        Reference:
        https://docs.sonos.com/reference/playbacksession-loadcloudqueue-sessionid
        """
        options = {
            "queueBaseUrl": queue_base_url,
        }
        if http_authorization is not None:
            options["httpAuthorization"] = http_authorization
        if use_http_authorization_for_media is not None:
            options["useHttpAuthorizationForMedia"] = use_http_authorization_for_media
        if item_id is not None:
            options["itemId"] = item_id
        if queue_version is not None:
            options["queueVersion"] = queue_version
        if position_millis is not None:
            options["positionMillis"] = position_millis
        if play_on_completion is not None:
            options["playOnCompletion"] = play_on_completion
        if track_metadata is not None:
            options["trackMetadata"] = track_metadata
        await self.api.send_command(
            namespace=self.namespace,
            command="loadCloudQueue",
            sessionId=session_id,
            options=options,
        )

    async def load_stream_url(
        self,
        session_id: str,
        stream_url: str,
        play_on_completion: bool | None = None,
        station_metadata: Container | None = None,
        item_id: str | None = None,
    ) -> None:
        """
        Send loadStreamUrl command to an (active) playback session).

        Reference:
        https://docs.sonos.com/reference/playbacksession-loadstreamurl-sessionid
        """
        options = {
            "streamUrl": stream_url,
        }
        if play_on_completion is not None:
            options["playOnCompletion"] = play_on_completion
        if station_metadata is not None:
            options["stationMetadata"] = station_metadata
        if item_id is not None:
            options["itemId"] = item_id
        await self.api.send_command(
            namespace=self.namespace,
            command="loadStreamUrl",
            sessionId=session_id,
            options=options,
        )

    async def refresh_cloud_queue(
        self,
        session_id: str,
    ) -> None:
        """
        Send refreshCloudQueue command to an (active) playback session).

        Reference:
        https://docs.sonos.com/reference/playbacksession-refreshcloudqueue-sessionid
        """
        await self.api.send_command(
            namespace=self.namespace,
            command="refreshCloudQueue",
            sessionId=session_id,
        )

    async def seek(
        self,
        session_id: str,
        position_millis: int,
        item_id: str | None = None,
    ) -> None:
        """
        Send seek command to an (active) playback session).

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playbacksession-seek-sessionid
        """
        options = {
            "positionMillis": position_millis,
        }
        if item_id is not None:
            options["itemId"] = item_id
        await self.api.send_command(
            namespace=self.namespace,
            command="seek",
            sessionId=session_id,
            options=options,
        )

    async def seek_relative(
        self,
        session_id: str,
        delta_millis: int,
        item_id: str | None = None,
    ) -> None:
        """
        Send seekRelative command to group.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playbacksession-seekrelative-sessionid
        """
        options = {
            "deltaMillis": delta_millis,
        }
        if item_id is not None:
            options["itemId"] = item_id
        await self.api.send_command(
            namespace=self.namespace,
            command="seekRelative",
            sessionId=session_id,
            options=options,
        )

    async def skip_to_item(
        self,
        session_id: str,
        item_id: str,
        queue_version: str | None = None,
        position_millis: int | None = None,
        play_on_completion: bool | None = None,
        track_metadata: Track | None = None,
    ) -> None:
        """
        Send skipToItem command to an (active) playback session).

        Reference:
        https://docs.sonos.com/reference/playbacksession-skiptoitem-sessionid
        """
        options = {
            "itemId": item_id,
        }
        if queue_version is not None:
            options["queueVersion"] = queue_version
        if position_millis is not None:
            options["positionMillis"] = position_millis
        if play_on_completion is not None:
            options["playOnCompletion"] = play_on_completion
        if track_metadata is not None:
            options["trackMetadata"] = track_metadata
        await self.api.send_command(
            namespace=self.namespace,
            command="skipToItem",
            sessionId=session_id,
            options=options,
        )

    async def suspend(
        self,
        session_id: str,
        queue_version: str | None = None,
    ) -> None:
        """
        Send suspend command to an (active) playback session).

        Reference:
        https://docs.sonos.com/reference/playbacksession-suspend-sessionid
        """
        options = {}
        if queue_version is not None:
            options["queueVersion"] = queue_version
        await self.api.send_command(
            namespace=self.namespace,
            command="suspend",
            sessionId=session_id,
            options=options,
        )

    async def subscribe(
        self,
        session_id: str,
        callback: SubscribeCallbackType,
    ) -> UnsubscribeCallbackType:
        """
        Subscribe to events in the playbackSession namespace for a specific sessionId.

        Returns handle to unsubscribe.

        Reference:
        https://docs.sonos.com/reference/playback-subscribe-groupid
        """
        return await self._handle_subscribe(session_id, callback)



"""
Sonos API Client to manage a single Sonos speaker, using the local websockets api.

Although the only connection method that is implemented,
is the local websockets connection, the cloud API shares the same models and
namespaces so this client could be extended to support the cloud API as well.
"""

from __future__ import annotations

import asyncio
from collections.abc import Callable
from typing import TYPE_CHECKING, Self

from aiosonos.const import EventType, GroupEvent, SonosEvent
from aiosonos.utils import get_discovery_info

from .api.websockets import SonosLocalWebSocketsApi
from .group import SonosGroup
from .player import SonosPlayer

if TYPE_CHECKING:
    from types import TracebackType

    from aiohttp import ClientSession

    from aiosonos.api.models import GroupInfo

    from .api.models import Group as GroupData
    from .api.models import Groups as GroupsData


EventCallBackType = Callable[[SonosEvent], None]
EventSubscriptionType = tuple[
    EventCallBackType,
    tuple[EventType, ...] | None,
    tuple[str, ...] | None,
]


class SonosLocalApiClient:
    """Sonos API Client to manage a single Sonos speaker, using the local websockets api."""

    api: SonosLocalWebSocketsApi
    _loop = asyncio.BaseEventLoop
    _household_id: str = ""
    _player_id: str
    _player: SonosPlayer | None = None

    def __init__(self, player_ip: str, aiohttp_session: ClientSession) -> None:
        """Initialize the Sonos API client."""
        self.player_ip = player_ip
        self._aiohttp_session = aiohttp_session
        self._groups: dict[str, SonosGroup] = {}
        self._subscribers: list[EventSubscriptionType] = []

    @property
    def player_id(self) -> str:
        """Return the player id."""
        return self._player_id

    @property
    def player(self) -> SonosPlayer:
        """Return the player."""
        return self._player

    @property
    def household_id(self) -> str:
        """Return the household id."""
        return self._household_id

    @property
    def groups(self) -> list[SonosGroup]:
        """Return all groups available in the Sonos household."""
        return list(self._groups.values())

    def get_group(self, group_id: str) -> SonosGroup:
        """Return a group."""
        return self._groups[group_id]

    def subscribe(
        self,
        cb_func: EventCallBackType,
        event_filter: EventType | tuple[EventType, ...] | None = None,
        object_id_filter: str | tuple[str] | None = None,
    ) -> Callable[[], None]:
        """Add callback to event listeners.

        Returns function to remove the listener.

        Parameters:
            - cb_func: callback function or coroutine
            - event_filter: Optionally only listen for these events
            - object_id_filter: Optionally only listen for these id's (player id, etc.)
        """
        if isinstance(event_filter, EventType):
            event_filter = (event_filter,)
        if isinstance(object_id_filter, str):
            object_id_filter = (object_id_filter,)
        listener = (cb_func, event_filter, object_id_filter)
        self._subscribers.append(listener)

        def remove_listener() -> None:
            self._subscribers.remove(listener)

        return remove_listener

    def signal_event(self, event: SonosEvent) -> None:
        """Forward event to subscribers."""
        for cb_func, event_filter, id_filter in self._subscribers:
            if not (event_filter is None or event.event_type in event_filter):
                continue
            if not (id_filter is None or event.object_id in id_filter):
                continue
            self._loop.call_soon_threadsafe(cb_func, event)

    async def connect(self) -> None:
        """Connect to the API."""
        self._loop = asyncio.get_running_loop()
        # retrieve discovery details from player first
        discovery_info = await get_discovery_info(self._aiohttp_session, self.player_ip)
        self._player_id = discovery_info["playerId"]
        self._household_id = discovery_info["householdId"]
        # Connect to the local websocket API
        self.api = SonosLocalWebSocketsApi(
            discovery_info["websocketUrl"],
            self._aiohttp_session,
        )
        # NOTE: connect will raise when connecting failed
        await self.api.connect()

    async def disconnect(self) -> None:
        """Disconnect the client and cleanup."""
        await self.api.disconnect()

    async def start_listening(self, init_ready: asyncio.Event | None = None) -> None:
        """Fetch initial state keep listening to messages until stopped."""
        listen_task = asyncio.create_task(self.api.start_listening())
        # fetch all initial data and setup subscriptions
        groups_data = await self.api.groups.get_groups(
            self.household_id,
            include_device_info=True,
        )
        for group_data in groups_data["groups"]:
            await self._setup_group(group_data)
        # Although all players are returned in the groups data,
        # the local api can only process the player that it is connected to
        # so we ignore all other player objects. For each Sonos player,
        # an individual api connection should be set-up to manage the player.
        # The Cloud API however is able to manage all players in the household.
        player_data = next(x for x in groups_data["players"] if x["id"] == self._player_id)
        self._player = player = SonosPlayer(self, player_data)
        await player.async_init()
        # setup global groups/player subscription
        await self.api.groups.subscribe(self._household_id, self._handle_groups_event)
        # set event that initial setup is done
        init_ready.set()
        # wait (forever) for incoming messages
        await listen_task

    async def create_group(
        self,
        player_ids: list[str],
        music_context_group_id: str | None = None,
    ) -> GroupInfo:
        """
        Create a new group.

        Use the createGroup command in the groups namespace to
        create a new group from a list of players. The player returns
        a group object with the group ID. This may be an existing
        group ID if an existing group is a subset of the new group.
        In this case, Sonos may build the new group by adding new
        players to the existing group.

        musicContextGroupId (Optional): The group containing the audio
        that you want to use. If empty or not provided,
        the new group will not contain any audio.
        """
        await self.api.groups.create_group(
            self._household_id,
            player_ids,
            music_context_group_id,
        )

    def _handle_groups_event(self, groups_data: GroupsData) -> None:
        """Handle a groups event."""
        # handle group updates
        for group_data in groups_data["groups"]:
            if group_data["id"] in self._groups:
                # existing group object
                self._groups[group_data["id"]].update_data(group_data)
                continue
            # a new group was added
            self._loop.create_task(self._setup_group(group_data))
        # check if any groups are removed
        removed_groups = set(self._groups.keys()) - {g["id"] for g in groups_data["groups"]}
        for group_id in removed_groups:
            group = self._groups.pop(group_id)
            self.signal_event(
                GroupEvent(
                    EventType.GROUP_REMOVED,
                    group_id,
                    group,
                ),
            )
        # handle player updates
        for player_data in groups_data["players"]:
            if player_data["id"] != self._player_id:
                continue
            self._player.update_data(player_data)

    async def _setup_group(self, group_data: GroupData) -> None:
        """Register/setup a (new) group."""
        group = SonosGroup(self, group_data)
        self._groups[group.id] = group
        await group.async_init()
        # always let the player check if the group changed,
        # as this might have been the active group
        if self._player:
            self._player.check_active_group()
        self.signal_event(
            GroupEvent(
                EventType.GROUP_ADDED,
                group.id,
                group,
            ),
        )

    async def __aenter__(self) -> Self:
        """Initialize and connect."""
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        """Exit and disconnect."""
        await self.disconnect()



"""Handle PlayerVolume related endpoints for Sonos."""

from __future__ import annotations

from aiosonos.api.models import PlayerVolume

from ._base import SonosNameSpace, SubscribeCallbackType, UnsubscribeCallbackType


class PlayerVolumeNameSpace(SonosNameSpace):
    """PlayerVolume Namespace handlers."""

    namespace = "playerVolume"
    event_type = "playerVolume"
    _event_model = PlayerVolume
    _event_key = "playerId"

    async def set_volume(
        self,
        player_id: str,
        volume: int | None = None,
        muted: bool | None = None,
    ) -> None:
        """
        Send setVolume command to set the player's volume and mute state.

        Note that when connected to a local speaker's websocket,
        the player id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playervolume-setvolume-playerid
        """
        options = {}
        if volume is not None:
            options["volume"] = volume
        if muted is not None:
            options["muted"] = muted
        await self.api.send_command(
            namespace=self.namespace,
            command="setVolume",
            playerId=player_id,
            options=options,
        )

    async def get_volume(
        self,
        player_id: str,
    ) -> PlayerVolume:
        """
        Get the player's volume and mute state.

        Note that when connected to a local speaker's websocket,
        the player id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playervolume-getvolume-playerid
        """
        return await self.api.send_command(
            namespace=self.namespace,
            command="getVolume",
            playerId=player_id,
        )

    async def duck(
        self,
        player_id: str,
        duration_millis: int | None = None,
    ) -> None:
        """
        Send volume duck command to player.

        Note that when connected to a local speaker's websocket,
        the player id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playervolume-duck-playerid
        """
        await self.api.send_command(
            namespace=self.namespace,
            command="duck",
            playerId=player_id,
            options={"durationMillis": duration_millis} if duration_millis else None,
        )

    async def set_mute(
        self,
        player_id: str,
        muted: bool,
    ) -> None:
        """
        Send the setMute command to set the player's mute state.

        Note that when connected to a local speaker's websocket,
        the player id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playervolume-setmute-playerid
        """
        await self.api.send_command(
            namespace=self.namespace,
            command="setMute",
            playerId=player_id,
            options={"muted": muted},
        )

    async def set_relative_volume(
        self,
        player_id: str,
        volume_delta: int | None = None,
        muted: bool | None = None,
    ) -> None:
        """
        Send setRelativeVolume command to set the player's volume and mute state.

        Note that when connected to a local speaker's websocket,
        the player id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playervolume-setvolume-playerid
        """
        options = {}
        if volume_delta is not None:
            options["volumeDelta"] = volume_delta
        if muted is not None:
            options["muted"] = muted
        await self.api.send_command(
            namespace=self.namespace,
            command="setRelativeVolume",
            playerId=player_id,
            options=options,
        )

    async def subscribe(
        self,
        player_id: str,
        callback: SubscribeCallbackType,
    ) -> UnsubscribeCallbackType:
        """
        Subscribe to events in the PlayerVolume namespace for given player.

        Returns handle to unsubscribe.

        Reference:
        https://docs.sonos.com/reference/playervolume-subscribe-playerid
        """
        return await self._handle_subscribe(player_id, callback)

    async def unduck(
        self,
        player_id: str,
    ) -> None:
        """
        Send volume unduck command to player.

        Reference:
        https://docs.sonos.com/reference/playervolume-unduck-playerid
        """
        await self.api.send_command(
            namespace=self.namespace,
            command="unduck",
            playerId=player_id,
        )



"""
Websocket API client for Sonos devices.

Simple wrapper for the Websocket API provided by Sonos which runs in the cloud,
but also locally on every Sonos speaker.

As far as I can oversee, the websockets API only runs locally on the Sonos speakers
and is not available in the cloud. The cloud API is a REST API and is documented here:
https://developer.sonos.com/reference

The objects, namespaces and commands however are all the same between the two APIs.
So at one point an alternative version of this client could be created
to support the cloud API as well.

In that case, commands are sent by regular HTTP GET/POST/PUT/DELETE requests,
and events received by a callback URL that you have to provide to the Sonos cloud API.
"""

from __future__ import annotations

import asyncio
import pprint
import uuid
from typing import TYPE_CHECKING, Any

import orjson
from aiohttp import ClientSession, ClientWebSocketResponse, WSMsgType, client_exceptions

from aiosonos.api.models import CommandMessage, ResultMessage
from aiosonos.const import LOCAL_API_TOKEN, LOG_LEVEL_VERBOSE
from aiosonos.exceptions import (
    CannotConnect,
    ConnectionClosed,
    ConnectionFailed,
    FailedCommand,
    InvalidMessage,
    InvalidState,
    NotConnected,
)
from aiosonos.utils import create_ssl_context

from ._base import AbstractSonosApi

if TYPE_CHECKING:
    from aiohttp import ClientSession

API_VERSION = 1


class SonosLocalWebSocketsApi(AbstractSonosApi):
    """Manage a Sonos Speaker using the local websockets api."""

    def __init__(
        self,
        websocket_url: str,
        aiohttp_session: ClientSession,
    ) -> None:
        """Initialize the Sonos API Connection to a local player's websocket."""
        super().__init__()
        self._aiohttp_session = aiohttp_session
        self.websocket_url = websocket_url
        self._ws_client: ClientWebSocketResponse | None = None
        self._result_futures: dict[str, asyncio.Future] = {}

    @property
    def connected(self) -> bool:
        """Return if we're currently connected."""
        return self._ws_client is not None and not self._ws_client.closed

    async def send_command(
        self,
        namespace: str,
        command: str,
        options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        """Send a command and get a response."""
        if not self.connected or not self._loop:
            raise InvalidState("Not connected")

        command_message = CommandMessage(
            namespace=f"{namespace}:{API_VERSION}",
            command=command,
            cmdId=uuid.uuid4().hex,
            # path params are passed as kwargs
            **kwargs,
        )
        future: asyncio.Future[Any] = self._loop.create_future()
        self._result_futures[command_message["cmdId"]] = future
        # body params are passed as options
        await self._send_message([command_message, options or {}])
        try:
            return await future
        finally:
            self._result_futures.pop(command_message["cmdId"])

    def send_command_no_wait(
        self,
        namespace: str,
        command: str,
        options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Send a command without waiting for the response."""
        if not self.connected:
            msg = "Not connected"
            raise InvalidState(msg)

        command_message = CommandMessage(
            namespace=namespace,
            command=command,
            cmdId=uuid.uuid4().hex,
            # path params are passed as kwargs
            **kwargs,
        )
        # body params are passed as options
        self.create_task(self._send_message([command_message, options or {}]))

    async def start_listening(self) -> None:
        """Connect (if needed) and start listening to incoming messages from the server."""
        if not self.connected:
            await self.connect()

        try:
            # keep reading incoming messages
            while not self._stop_called:
                msg = await self.receive_message_or_raise()
                self._handle_incoming_message(msg)
        except ConnectionClosed:
            pass
        finally:
            await self.disconnect()

    async def connect(self) -> None:
        """Connect to the websocket server."""
        self._loop = asyncio.get_running_loop()
        if self._ws_client is not None:
            raise InvalidState("Already connected")

        self.logger.debug("Trying to connect to %s", self.websocket_url)

        headers = {
            "X-Sonos-Api-Key": LOCAL_API_TOKEN,
            "Sec-Websocket-Protocol": "v1.api.smartspeaker.audio",
        }
        try:
            self._ws_client = await self._aiohttp_session.ws_connect(
                self.websocket_url,
                heartbeat=55,
                compress=15,
                max_msg_size=0,
                headers=headers,
                ssl=create_ssl_context(),
            )
        except (
            client_exceptions.WSServerHandshakeError,
            client_exceptions.ClientError,
        ) as err:
            raise CannotConnect(err) from err

        self.logger.info(
            "Connected to Sonos Websocket (%s)",
            self.websocket_url,
        )

    async def disconnect(self) -> None:
        """Disconnect the client and cleanup."""
        self._stop_called = True
        # cancel all command-tasks awaiting a result
        for future in self._result_futures.values():
            future.cancel()
        self.logger.debug("Closing client connection")
        if self._ws_client is not None and not self._ws_client.closed:
            await self._ws_client.close()
        self._ws_client = None

    def _handle_incoming_message(
        self,
        raw: tuple[ResultMessage, dict[str, Any]],
    ) -> None:
        """
        Handle incoming message.

        Run all async tasks in a wrapper to log appropriately.
        """
        msg, msg_data = raw

        # handle error message
        if "errorCode" in msg_data:
            if "cmdId" not in msg:
                self.logger.error("Received unhandled error: %s: %s", msg, msg_data)
                return
            if future := self._result_futures.get(msg["cmdId"]):
                future.set_exception(
                    FailedCommand(msg_data["errorCode"], msg_data.get("reason")),
                )
            return

        # handle command result message
        if "success" in msg:
            if future := self._result_futures.get(msg["cmdId"]):
                if msg["success"]:
                    future.set_result(msg_data)
                else:
                    future.set_exception(FailedCommand(msg_data["_objectType"]))
            return

        # handle EventMessage
        if event_type := msg.get("type"):
            # handle namespace specific events
            for namespace in (
                self._audio_clip,
                self._groups,
                self._group_volume,
                self._playback_metadata,
                self._playback_session,
                self._playback,
                self._player_volume,
            ):
                if event_type == namespace.event_type:
                    self.create_task(namespace._handle_event(msg, msg_data))  # noqa: SLF001
                    break
            else:
                self.logger.debug(
                    "Received unhandled event type: %s: %s",
                    event_type,
                    msg,
                )
            return

        # Log anything we can't handle here
        self.logger.debug("Received unhandled message: %s", msg)

    async def receive_message_or_raise(self) -> tuple[ResultMessage, dict[str, Any]]:
        """Receive (raw) message or raise."""
        assert self._ws_client
        ws_msg = await self._ws_client.receive()

        if ws_msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSED, WSMsgType.CLOSING):
            raise ConnectionClosed("Connection was closed.")

        if ws_msg.type == WSMsgType.ERROR:
            raise ConnectionFailed

        if ws_msg.type != WSMsgType.TEXT:
            err_msg = f"Received non-Text message: {ws_msg.type}: {ws_msg.data}"
            raise InvalidMessage(err_msg)

        try:
            msg = orjson.loads(ws_msg.data)
        except TypeError as err:
            err_msg = f"Received unsupported JSON: {err}"
            raise InvalidMessage(msg) from err_msg
        except ValueError as err:
            raise InvalidMessage("Received invalid JSON.") from err

        if self.logger.isEnabledFor(LOG_LEVEL_VERBOSE):
            self.logger.log(
                LOG_LEVEL_VERBOSE,
                "Received message:\n%s\n",
                pprint.pformat(ws_msg),
            )

        return msg

    async def _send_message(
        self,
        message: tuple[CommandMessage, dict[str, Any]],
    ) -> None:
        """
        Send a message to the server.

        Raises NotConnected if client not connected.
        """
        if not self.connected:
            raise NotConnected

        if self.logger.isEnabledFor(LOG_LEVEL_VERBOSE):
            self.logger.log(
                LOG_LEVEL_VERBOSE,
                "Publishing message:\n%s\n",
                pprint.pformat(message),
            )

        assert self._ws_client
        # sonos messages are always an array of 2 dict objects
        assert isinstance(message, list)

        await self._ws_client.send_bytes(orjson.dumps(message))



"""Base/Model for namespace-specific Sonos controller."""

from __future__ import annotations

import logging
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from aiosonos.api import SonosLocalWebSocketsApi

_T = TypeVar("_T")

SubscribeCallbackType = Callable[[_T], None]
UnsubscribeCallbackType = Callable[[], None]

_LOGGER = logging.getLogger(__name__)


class SonosNameSpace:
    """Base/Model for namespace-specific Sonos controller."""

    namespace: str
    event_type: str
    _event_model: _T
    _event_key: str

    def __init__(self, api: SonosLocalWebSocketsApi) -> None:
        """Handle Initialization."""
        self.api = api
        self._listeners: dict[str, SubscribeCallbackType] = {}

    async def _handle_event(self, event: dict[str, Any], event_data: dict[str, Any]) -> None:
        """Handle incoming event from subscription."""
        event_id = event[self._event_key]
        if handler := self._listeners.get(event_id):
            handler(event_data)

    async def _handle_subscribe(
        self,
        event_id: str,
        callback: SubscribeCallbackType,
    ) -> UnsubscribeCallbackType:
        """Handle subscription logic."""
        if event_id in self._listeners:
            _LOGGER.error("Duplicate subscription detected for %s", event_id)
        await self.api.send_command(
            namespace=self.namespace,
            command="subscribe",
            **{self._event_key: event_id},
        )
        self._listeners[event_id] = callback

        def _unsubscribe() -> None:
            self._listeners.pop(event_id)
            self.api.send_command_no_wait(
                namespace=self.namespace,
                command="unsubscribe",
                **{self._event_key: event_id},
            )

        return _unsubscribe



"""Handle PlaybackMetadata related endpoints for Sonos."""

from __future__ import annotations

from aiosonos.api.models import MetadataStatus

from ._base import SonosNameSpace, SubscribeCallbackType, UnsubscribeCallbackType


class PlaybackMetadataNameSpace(SonosNameSpace):
    """PlaybackMetadataNameSpace Namespace handlers."""

    namespace = "playbackMetadata"
    event_type = "metadataStatus"
    _event_model = MetadataStatus
    _event_key = "groupId"

    async def get_metadata_status(
        self,
        group_id: str,
    ) -> MetadataStatus:
        """
        Send getMetadataStatus command to group.

        Note that when connected to a local speaker's websocket,
        the group id can only be that from the local speaker itself.

        Reference:
        https://docs.sonos.com/reference/playbackmetadata-getmetadatastatus-groupid
        """
        return await self.api.send_command(
            namespace=self.namespace,
            command="getMetadataStatus",
            groupId=group_id,
        )

    async def subscribe(
        self,
        group_id: str,
        callback: SubscribeCallbackType,
    ) -> UnsubscribeCallbackType:
        """
        Subscribe to events in the MetadataStatus namespace for given player.

        Returns handle to unsubscribe.

        Reference:
        https://docs.sonos.com/reference/playback-subscribe-groupid
        """
        return await self._handle_subscribe(group_id, callback)



"""Base model for a Sonos API Controller."""

from __future__ import annotations

import logging
import uuid
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Self

from aiosonos.api.namespaces.audio_clip import AudioClipNameSpace
from aiosonos.api.namespaces.group_volume import GroupVolumeNameSpace
from aiosonos.api.namespaces.groups import GroupsNameSpace
from aiosonos.api.namespaces.home_theater import HomeTheaterNameSpace
from aiosonos.api.namespaces.playback import PlaybackNameSpace
from aiosonos.api.namespaces.playback_metadata import PlaybackMetadataNameSpace
from aiosonos.api.namespaces.playback_session import PlaybackSessionNameSpace
from aiosonos.api.namespaces.player_volume import PlayerVolumeNameSpace

if TYPE_CHECKING:
    import asyncio
    from collections.abc import Awaitable
    from types import TracebackType


class AbstractSonosApi:
    """Base abstraction for a Sonos API."""

    def __init__(
        self,
    ) -> None:
        """Initialize."""
        self.logger = logging.getLogger(__package__)
        self._stop_called: bool = False
        self._loop: asyncio.AbstractEventLoop | None = None
        self._audio_clip = AudioClipNameSpace(self)
        self._groups = GroupsNameSpace(self)
        self._group_volume = GroupVolumeNameSpace(self)
        self._playback = PlaybackNameSpace(self)
        self._playback_metadata = PlaybackMetadataNameSpace(self)
        self._playback_session = PlaybackSessionNameSpace(self)
        self._player_volume = PlayerVolumeNameSpace(self)
        self._home_theater = HomeTheaterNameSpace(self)
        self._tracked_tasks: dict[str, asyncio.Task] = {}

    @property
    @abstractmethod
    def connected(self) -> bool:
        """Return if we're currently connected."""

    @property
    def audio_clip(self) -> AudioClipNameSpace:
        """Return AudioClip namespace handler."""
        return self._audio_clip

    @property
    def groups(self) -> GroupsNameSpace:
        """Return Groups namespace handler."""
        return self._groups

    @property
    def group_volume(self) -> GroupVolumeNameSpace:
        """Return GroupVolume namespace handler."""
        return self._group_volume

    @property
    def playback(self) -> PlaybackNameSpace:
        """Return PlayBack namespace handler."""
        return self._playback

    @property
    def playback_metadata(self) -> PlaybackMetadataNameSpace:
        """Return PlaybackMetadata namespace handler."""
        return self._playback_metadata

    @property
    def playback_session(self) -> PlaybackSessionNameSpace:
        """Return PlaybackSession namespace handler."""
        return self._playback_session

    @property
    def player_volume(self) -> PlayerVolumeNameSpace:
        """Return PlayerVolume namespace handler."""
        return self._player_volume

    @property
    def home_theater(self) -> HomeTheaterNameSpace:
        """Return HomeTheater namespace handler."""
        return self._home_theater

    @abstractmethod
    async def send_command(
        self,
        namespace: str,
        command: str,
        options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        """Send a command and get a response."""

    @abstractmethod
    def send_command_no_wait(
        self,
        namespace: str,
        command: str,
        options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Send a command without waiting for the response."""

    @abstractmethod
    async def start_listening(self, init_ready: asyncio.Event | None = None) -> None:
        """Connect (if needed) and start listening to incoming messages from the server."""

    @abstractmethod
    async def connect(self) -> None:
        """Connect to the server/api."""

    @abstractmethod
    async def disconnect(self) -> None:
        """Disconnect the client and cleanup."""

    def create_task(
        self,
        target: Awaitable,
    ) -> asyncio.Task:
        """
        Create Task on (main) event loop from Coroutine(function).

        Tasks created by this helper will be properly cancelled on stop,
        and exceptions will be logged.
        """

        def task_done_callback(_task: asyncio.Future | asyncio.Task) -> None:
            _task_id = task.task_id
            self._tracked_tasks.pop(_task_id)
            # log unhandled exceptions
            if not _task.cancelled() and (err := _task.exception()):
                task_name = _task.get_name() if hasattr(_task, "get_name") else str(_task)
                self.logger.warning(
                    "Exception in task %s - target: %s: %s",
                    task_name,
                    str(target),
                    str(err),
                    exc_info=err if self.logger.isEnabledFor(logging.DEBUG) else None,
                )

        task = self._loop.create_task(target)
        task_id = uuid.uuid4().hex
        task.task_id = task_id
        self._tracked_tasks[task_id] = task
        task.add_done_callback(task_done_callback)
        return task

    async def __aenter__(self) -> Self:
        """Initialize and connect the connection to the SonosApi."""
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        """Exit context manager."""
        await self.disconnect()

    def __repr__(self) -> str:
        """Return the representation."""
        conn_type = self.__class__.__name__
        prefix = "" if self.connected else "not "
        return f"{type(self).__name__}(connection={conn_type}, {prefix}connected)"



"""
Models/schemas for objects in the Sonos HTTP/Websockets API.

Reference: https://docs.sonos.com/docs/types
"""

from __future__ import annotations

from enum import StrEnum
from typing import NotRequired, TypedDict


class CommandMessage(TypedDict):
    """Representation of a Command message."""

    namespace: str  # e.g. 'groups'
    command: str  # e.g. 'getGroups'
    sessionId: NotRequired[str]  # optional sessionId parameter to pass along
    cmdId: NotRequired[str]  # optional cmdId to pass along
    # other command specific parameters are just added to this dict


class ErrorResponse(TypedDict):
    """Representation of an error response message."""

    errorCode: str  # e.g. '401'
    reason: NotRequired[str]  # e.g. 'ERROR_INVALID_PARAMETER'


class ResultMessage(TypedDict):
    """Base Representation of a message received from the api/server to the client."""

    namespace: str  # the namespace the command was directed to
    response: str  # the command that was executed
    householdId: str  # housholdId is always returned in a response
    type: str  # an optional response type
    sessionId: NotRequired[str]  # optional sessionId parameter to pass along
    cmdId: NotRequired[str]  # optional cmdId to pass along
    success: NotRequired[bool]  # in case of a command response


class AudioClipPriority(StrEnum):
    """
    Enum with possible AudioClip priorities.

    Reference: https://docs.sonos.com/reference/audioclip-loadaudioclip-playerid
    """

    LOW = "LOW"
    HIGH = "HIGH"


class AudioClipType(StrEnum):
    """
    Enum with possible AudioClip types.

    Reference: https://docs.sonos.com/reference/audioclip-loadaudioclip-playerid
    """

    CHIME = "CHIME"
    CUSTOM = "CUSTOM"
    VOICE_ASSISTANT = "VOICE_ASSISTANT"


class AudioClipLEDBehavior(StrEnum):
    """
    Enum with possible LEDBehavior behaviors.

    Reference: https://docs.sonos.com/reference/audioclip-loadaudioclip-playerid
    """

    WHITE_LED_QUICK_BREATHING = "WHITE_LED_QUICK_BREATHING"
    NONE = "NONE"


class AudioClipStatus(StrEnum):
    """
    Enum with possible LEDBehavior behaviors.

    Reference: https://docs.sonos.com/reference/audioclip-loadaudioclip-playerid
    """

    ACTIVE = "ACTIVE"
    DONE = "DONE"
    DISMISSED = "DISMISSED"
    INACTIVE = "INACTIVE"
    INTERRUPTED = "INTERRUPTED"
    ERROR = "ERROR"


class AudioClip(TypedDict):
    """
    Representation of an AudioClip object/response.

    Response message from loadAudioClip command.
    Reference: https://docs.sonos.com/reference/audioclip-loadaudioclip-playerid
    """

    _objectType: str  # = audioClip
    id: str  # The unique identifier for the audio clip.
    name: str  # User identifiable string.
    appId: str  # The unique identifier for the app that created the audio clip.
    priority: AudioClipPriority
    clipType: AudioClipType  # e.g. 'CHIME'
    status: AudioClipStatus  # This field indicates the state of the audio clip
    clipLEDBehavior: AudioClipLEDBehavior  # e.g. 'WHITE_LED_QUICK_BREATHING'


class AudioClipStatusEvent(TypedDict):
    """
    Representation of an AudioClipStatus message, as received in events.

    Reference: https://docs.sonos.com/reference/audioclip-subscribe-playerid
    """

    _objectType: str  # = audioClipStatus
    audioClips: list[AudioClip]


class PlayerVolume(TypedDict):
    """
    Representation of a PlayerVolume object/response.

    Reference: https://docs.sonos.com/docs/types#playervolume
    """

    _objectType: str  # = playerVolume
    fixed: bool
    muted: bool
    volume: int


class GroupVolume(TypedDict):
    """
    Representation of a GroupVolume object/response.

    Reference: https://docs.sonos.com/docs/types#groupvolume
    """

    _objectType: str  # = groupVolume
    fixed: bool
    muted: bool
    volume: int


class PlayBackState(StrEnum):
    """Enum with possible playback states."""

    PLAYBACK_STATE_IDLE = "PLAYBACK_STATE_IDLE"
    PLAYBACK_STATE_BUFFERING = "PLAYBACK_STATE_BUFFERING"
    PLAYBACK_STATE_PAUSED = "PLAYBACK_STATE_PAUSED"
    PLAYBACK_STATE_PLAYING = "PLAYBACK_STATE_PLAYING"


class SonosCapability(StrEnum):
    """Enum with possible Sonos (device) capabilities."""

    CLOUD = "CLOUD"
    PLAYBACK = "PLAYBACK"
    AIRPLAY = "AIRPLAY"
    LINE_IN = "LINE_IN"
    VOICE = "VOICE"
    AUDIO_CLIP = "AUDIO_CLIP"
    MICROPHONE_SWITCH = "MICROPHONE_SWITCH"
    HT_PLAYBACK = "HT_PLAYBACK"


class Group(TypedDict):
    """Representation of a group."""

    _objectType: str  # = group
    coordinatorId: str
    id: str
    name: str
    playbackState: NotRequired[str]
    playerIds: list[str]
    areaIds: NotRequired[list[str]]


class GroupInfo(TypedDict):
    """Representation of a GroupInfo object/event."""

    _objectType: str  # = groupInfo
    group: Group


class Player(TypedDict):
    """Representation of a player."""

    _objectType: str  # = player
    id: str
    name: str
    websocketUrl: str
    softwareVersion: str
    apiVersion: str
    minApiVersion: str
    devices: list[DeviceInfo]
    zoneInfo: ActiveZone


class DeviceInfo(TypedDict):
    """Representation of device information."""

    _objectType: str  # = deviceInfo
    id: str
    primaryDeviceId: NotRequired[str]
    serialNumber: NotRequired[str]
    modelDisplayName: NotRequired[str]
    color: NotRequired[str]
    capabilities: list[SonosCapability]
    apiVersion: NotRequired[str]
    minApiVersion: NotRequired[str]
    name: NotRequired[str]
    websocketUrl: NotRequired[str]
    softwareVersion: NotRequired[str]
    hwVersion: NotRequired[str]
    swGen: NotRequired[int]


class ZoneMemberState(TypedDict):
    """Representation of a zone member state."""

    _objectType: str  # = zoneMemberState
    disconnected: bool


class ActiveZoneMember(TypedDict):
    """Representation of an active zone member."""

    _objectType: str  # = activeZoneMember
    channelMap: list[str]
    id: str
    state: ZoneMemberState


class ActiveZone(TypedDict):
    """Representation of an active zone."""

    _objectType: str  # = activeZone
    members: list[ActiveZoneMember]
    name: str
    zoneId: NotRequired[str]


class Groups(TypedDict):
    """
    Representation of a Groups message (event or response).

    Reference: https://docs.sonos.com/docs/types#groups
    """

    _objectType: str  # = groups
    groups: list[Group]
    partial: bool
    players: list[Player]


class DiscoveryInfo(TypedDict):
    """Representation of discoveryInfo."""

    _objectType: str  # = discoveryInfo
    device: DeviceInfo
    householdId: str
    locationId: str
    playerId: str
    groupId: str
    websocketUrl: str
    restUrl: str


class PlaybackActions(TypedDict):
    """Representation of available playback actions."""

    _objectType: str  # = playbackAction
    canCrossfade: bool
    canPause: bool
    canPlay: bool
    canRepeat: bool
    canRepeatOne: bool
    canSeek: bool
    canShuffle: bool
    canSkip: bool
    canSkipBack: bool
    canSkipToPrevious: bool
    canStop: bool


class PlayModes(TypedDict):
    """Representation of play modes."""

    _objectType: str  # = playModes
    crossfade: NotRequired[bool]
    repeat: NotRequired[bool]
    repeatOne: NotRequired[bool]
    shuffle: NotRequired[bool]


class PlaybackStatus(TypedDict):
    """Representation of a playback status."""

    _objectType: str  # = playbackStatus
    availablePlaybackActions: PlaybackActions
    isDucking: bool
    playbackState: PlayBackState
    playModes: PlayModes
    positionMillis: int
    previousPositionMillis: int


class MusicService(StrEnum):
    """Enum with (known) possible container service Id's."""

    SPOTIFY = "9"
    MUSIC_ASSISTANT = "mass"
    TUNEIN = "303"
    QOBUZ = "31"
    YOUTUBE_MUSIC = "284"
    LOCAL_LIBRARY = "local-library"
    # TODO: complete this list with other known services


class MetadataId(TypedDict):
    """Representation of an ID, used in metadata objects."""

    _objectType: str  # = id
    serviceId: NotRequired[str]
    objectId: NotRequired[str]
    accountId: NotRequired[str]


class Service(TypedDict):
    """Representation of a service."""

    _objectType: str  # = service
    name: str
    images: NotRequired[list[Image]]


class Image(TypedDict):
    """Representation of an image."""

    _objectType: str  # = image
    url: str
    name: NotRequired[str]
    type: NotRequired[str]


class Author(TypedDict):
    """Representation of an author."""

    name: str
    id: MetadataId


class Narrator(TypedDict):
    """Representation of a narrator."""

    name: str
    id: MetadataId


class Producer(TypedDict):
    """Representation of a podcast producer."""

    name: str
    id: NotRequired[MetadataId]


class Book(TypedDict):
    """Representation of a (audio)book."""

    _objectType: str  # = book
    name: str
    chapterCount: NotRequired[int]
    author: NotRequired[Author]
    narrator: NotRequired[Narrator]
    id: NotRequired[MetadataId]


class Podcast(TypedDict):
    """Representation of a podcast."""

    name: str
    producer: NotRequired[Producer]
    id: NotRequired[MetadataId]
    explicit: NotRequired[bool]


class Album(TypedDict):
    """Representation of an album."""

    _objectType: str  # = album

    name: str


class Artist(TypedDict):
    """Representation of an artist."""

    _objectType: str  # = artist

    name: str


class Quality(TypedDict):
    """Representation of track quality."""

    _objectType: str  # = quality

    bitDepth: NotRequired[int]
    sampleRate: NotRequired[int]
    codec: NotRequired[str]
    lossless: NotRequired[bool]
    immersive: NotRequired[bool]
    replayGain: NotRequired[float]


class Track(TypedDict):
    """Representation of a track."""

    _objectType: str  # = track

    type: str
    name: str
    mediaUrl: NotRequired[str]
    images: NotRequired[list[Image]]
    contentType: NotRequired[str]
    album: NotRequired[Album]
    artist: NotRequired[Artist]
    author: NotRequired[Author]
    book: NotRequired[Book]
    narrator: NotRequired[Narrator]
    podcast: NotRequired[Podcast]
    producer: NotRequired[Producer]
    releaseDate: NotRequired[str]
    episodeNumber: NotRequired[int]
    id: NotRequired[MetadataId]
    service: NotRequired[Service]
    durationMillis: NotRequired[int]
    trackNumber: NotRequired[int]
    chapterNumber: NotRequired[int]
    explicit: NotRequired[bool]


class Show(TypedDict):
    """
    Representation of a (Radio)Show).

    Information about the current "show", when available.
    Generally only present for radio stations (container.type = "station")
    """

    _objectType: str  # = show
    name: str
    id: NotRequired[MetadataId]
    images: NotRequired[list[Image]]
    explicit: NotRequired[bool]


class PlaybackSession(TypedDict):
    """
    Representation of a PlaybackSession.

    Provides details on the external source that initiated the playback.
    """

    _objectType: str  # = show
    clientId: str
    isSuspended: bool
    accountId: str


class ContainerType(StrEnum):
    """Enum with possible container types."""

    LINEIN = "linein"
    STATION = "station"
    PLAYLIST = "playlist"
    AIRPLAY = "linein.airplay"
    PODCAST = "podcast"
    BOOK = "book"
    ARTIST = "artist"
    ALBUM = "album"
    ARTIST_LOCAL = "artist.local"
    ALBUM_LOCAL = "album.local"
    HOME_THEATER_SPDIF = "linein.homeTheater.spdif"
    HOME_THEATER_HDMI = "linein.homeTheater.hdmi"

    # TODO: complete this list with other known types


class Container(TypedDict):
    """
    Representation of a container.

    A container object indicating the current playback source.
    The container describes and identifies what is currently playing,
    for example, the programmed radio station, music service playlist,
    or linein source.

    If no content is loaded, the container field will not be present.
    """

    _objectType: str  # = container
    name: str
    type: str
    id: NotRequired[MetadataId]
    service: NotRequired[Service]
    images: NotRequired[list[Image]]
    book: NotRequired[Book]
    podcast: NotRequired[Podcast]
    explicit: NotRequired[bool]


class QueueItem(TypedDict):
    """Representation of a Queue item."""

    _objectType: str  # = queueItem

    class Policies(TypedDict):
        """Representation of (playback)policies."""

        canSkip: NotRequired[bool]
        canSkipToPrevious: NotRequired[bool]
        limitedSkips: NotRequired[bool]
        canSeek: NotRequired[bool]
        canSkipToItem: NotRequired[bool]
        canRepeat: NotRequired[bool]
        canRepeatOne: NotRequired[bool]
        canCrossfade: NotRequired[bool]
        canShuffle: NotRequired[bool]
        canResume: NotRequired[bool]
        pauseAtEndOfQueue: NotRequired[bool]
        refreshAuthWhilePaused: NotRequired[bool]
        showNNextTracks: NotRequired[int]
        showNPreviousTracks: NotRequired[int]
        isVisible: NotRequired[bool]
        notifyUserIntent: NotRequired[bool]
        pauseTtlSec: NotRequired[int]
        playTtlSec: NotRequired[int]
        pauseOnDuck: NotRequired[bool]

    id: str
    deleted: NotRequired[bool]
    track: NotRequired[Track]
    policies: NotRequired[Policies]


class MetadataStatus(TypedDict):
    """Representation of (playback) metadata status."""

    _objectType: str  # = metadataStatus

    container: NotRequired[Container]
    currentItem: NotRequired[QueueItem]
    nextItem: NotRequired[QueueItem]
    currentShow: NotRequired[Show]
    streamInfo: NotRequired[str]
    playbackSession: NotRequired[PlaybackSession]


class SessionStatus(TypedDict):
    """Representation of a (Playback) session status."""

    _objectType: str  # = sessionStatus

    sessionId: str
    sessionState: str  # SESSION_STATE_CONNECTED
    sessionCreated: bool
    customData: str


class LoadContentRequest(TypedDict):
    """
    Representation of a playback content request.

    Sent when requesting the player(queue) to load content.

    example load spotify playlist:
    {
        "type": "PLAYLIST",
        "id": {
            "objectId": "spotify:playlist:5lH9NjOeJvctAO92ZrKQNB",
            "accountId": "sn_3",
            "serviceId": "9"
        },
        "playbackAction": "PLAY",
        "playModes": {
            "shuffle": false
        }
    }
    """

    type: str
    id: MetadataId
    playbackAction: str
    playModes: NotRequired[PlayModes]



"""
Manage a Sonos Topology/Household using HTTP or Websockets API.

Simple wrapper for the Websocket API provided by Sonos which runs in the cloud,
but also locally on every Sonos speaker.

API Reference can be found here:
https://docs.sonos.com/reference
"""

from ._base import AbstractSonosApi  # noqa: F401
from .websockets import SonosLocalWebSocketsApi  # noqa: F401
